proc TruncToResolution { value resolution } {
    set ratio [expr $value/$resolution]
    if { $ratio < 0 } {
	set resolution [expr $resolution * -1] 
	set ratio [expr $ratio * -1]
    }

    return [expr round($ratio) * $resolution]
}

######################################################################
# multiscalePixelToValue
#   Parameters:
#     this     - associative array for the multiscale
#     pixel    - pixel position
#
#   Actions:
#     Computes the value on the scale for the pixel position
#
#   Returns:
#     value on the scale for the pixel position
#
proc multiscalePixelToValue { this pixel } {
    
    global $this
    set widget [subst $${this}(widget)]

    set min_pixel [lindex [$widget coords rail] 0]
    set max_pixel [expr [lindex [$widget coords rail] 2] - [subst $${this}(rect_width)]]

    set min_value [subst $${this}(from)]
    set max_value [subst $${this}(to)]

    return [expr $min_value + ($max_value - $min_value) * ($pixel - $min_pixel) / ($max_pixel - $min_pixel)]
}

######################################################################
# multiscaleValueToPixel
#   Parameters:
#     this     - associative array for the multiscale
#     value    - value on the scale
#
#   Actions:
#     Computes the pixel position from the value on the scale
#
#   Returns:
#     pixel position
#
proc multiscaleValueToPixel { this value } {

    global $this
    set widget [subst $${this}(widget)]

    set min_pixel [lindex [$widget coords rail] 0]
    set max_pixel [expr [lindex [$widget coords rail] 2] - [subst $${this}(rect_width)]]

    set min_value [subst $${this}(from)]
    set max_value [subst $${this}(to)]

    return [expr $min_pixel + ($max_pixel - $min_pixel) * ($value - $min_value) / ($max_value - $min_value)]
}

######################################################################
# multiscaleVariableChanged
#   Parameters:
#     this     - associative array for the multiscale
#     index    - index of the slider
#
#   Actions:
#     Updates the graphics of the multiscale when a slider value changes
#
#   Returns:
#     nothing
#
proc multiscaleVariableChanged { this index var element op_type } {
    
    # This is the procedure which handles variable tracing

    global $this
    set widget [subst $[subst $this](widget)]

    global $var
    if { $element != {} } {
	set var ${var}($element)
    }
    
    set tag slider_$index
    
    # Find out where the box is currently
    set u_curr_pixel [lindex [$widget coords ${tag}_rect] 0]
    set u_new_pixel [multiscaleValueToPixel $this [subst $$var]]
    
    set du_pixel [expr $u_new_pixel - $u_curr_pixel]
    
    # Move the rect
    $widget move $tag $du_pixel 0
    
    # Update the hanging text
    set $var [multiscalePixelToValue $this [lindex [$widget coords ${tag}_rect] 0]]
    $widget itemconfigure ${tag}_text -text [TruncToResolution [subst $$var] [subst $${this}(resolution)]]
}

######################################################################
# multiscaleClick
#   Parameters:
#     this     - associative array for the multiscale
#     index    - index of the slider (unused)
#     u_pixel  - x pixel position of the click
#
#   Actions:
#     stores the starting click position
#
#   Returns:
#     nothing
#
proc multiscaleClick { this index u_pixel } {
    global $this

    # Remember the anchor point for the drag
    set ${this}(u_pixel) $u_pixel
}

######################################################################
# multiscaleDrag
#   Parameters:
#     this     - associative array for the multiscale
#     index    - index of the slider (unused)
#     u_pixel  - x pixel position of the click
#
#   Actions:
#     drags the slider and updates the variable
#
#   Returns:
#     nothing
#
proc multiscaleSet { var value } {
    set varBase ""
    regsub {\([_a-zA-Z0-9]*\)} $var "" varBase
    global $varBase

    set $var $value
}


proc multiscaleDrag { this index u_pixel } {

    global $this

    set widget [subst $${this}(widget)]
    set tag slider_$index
    set var [lindex [subst $${this}(variables)] $index]

    set varBase ""
    regsub {\([_a-zA-Z0-9]*\)} $var "" varBase
    global $varBase

    # Compute delta and update anchor point
    set u_last_pixel [subst $${this}(u_pixel)]
    set ${this}(u_pixel) $u_pixel
    set du_pixel [expr $u_pixel - $u_last_pixel]

    # Find out where the box is currently
    set u_curr_pixel [lindex [$widget coords ${tag}_rect] 0]

    # Find out what the constraints on the selected box are
    set u_min_pixel [lindex [$widget coords rail] 0]
    set u_max_pixel [expr [lindex [$widget coords rail] 2] - [subst $${this}(rect_width)]]

    if { $index > 0 } {
	set u_min_pixel [expr [lindex [$widget coords slider_[expr $index - 1]_rect] 0] + ([multiscaleValueToPixel $this [subst $${this}(spacing)]] - [multiscaleValueToPixel $this 0])]
    }
    if { [expr $index + 1] < [subst $${this}(sliders)] } {
  	set u_max_pixel [expr [lindex [$widget coords slider_[expr $index + 1]_rect] 0] - ([multiscaleValueToPixel $this [subst $${this}(spacing)]] - [multiscaleValueToPixel $this 0])]
    }

    # Move the current box as much as possible
    if { $u_curr_pixel + $du_pixel < $u_min_pixel && $u_curr_pixel + $du_pixel > $u_max_pixel } {
  	set du_pixel [expr (($u_max_pixel + $u_max_pixel) / 2) - $u_curr_pixel]
  	set ${this}(u_pixel) [expr ($u_max_pixel + $u_max_pixel) / 2]
    } elseif { $u_curr_pixel + $du_pixel < $u_min_pixel } {
	set du_pixel [expr $u_min_pixel - $u_curr_pixel]
	set ${this}(u_pixel) $u_min_pixel
    } elseif { $u_curr_pixel + $du_pixel > $u_max_pixel } {
	set du_pixel [expr $u_max_pixel - $u_curr_pixel]
	set ${this}(u_pixel) $u_max_pixel
    }
    $widget move $tag $du_pixel 0

    # Update the variable value and the hanging text
    set $var [multiscalePixelToValue $this [lindex [$widget coords ${tag}_rect] 0]]

    $widget itemconfigure ${tag}_text -text [TruncToResolution [subst $$var] [subst $${this}(resolution)]]

    if { $index < [llength [subst $${this}(commands)]] } {
	set com [lindex [subst $${this}(commands)] $index]
	eval $com $index [subst $$var]
    }
}


######################################################################
# multiscale 
#   Parameters:
#     name   - name of the radio button frame widget (convention: rbf<name>)
#     parent   - parent widget name
#     name     - name of the radio button frame widget (convention: rbf<name>)
#     var      - name of the global variable which will be modified
#     text     - text name to be displayed on the widget
#     enumList - enumerated list of fields and values { {field0 value0} ... }
#
#   Actions:
#     Creates a scale with multiple slider buttons
#
#   Returns:
#     0 if no error
#

set multiscaleParams [list \
	{ label { string } } \
	{ from { float } } \
	{ to { float } } \
	{ spacing { float } } \
	{ resolution { float } } \
	{ tickint { float } } \
	{ orient { string } } \
	{ length { float } } \
	{ variables { list } } \
	{ commands { list } } \
	]

set multiscalePrivates [list \
	{ sliders } \
	{ rect_width } \
	{ rect_height } \
	]

proc multiscale { name args } {

    ##################################################
    # Create an associative array object 
    # Process command line arguments and initialize the fields 

    global multiscaleParams

    # Create the associative array

    set widget $name
    set this ""
    regsub -all "\\." multiscaleobject$name "_" this
    
    global $this

    # Public fields
    set ${this}(label) ""
    set ${this}(from) 0.0
    set ${this}(to) 1.0
    set ${this}(spacing) 0.01 
    set ${this}(resolution) 0.01 
    set ${this}(tickint) ""
    set ${this}(variables) {}
    set ${this}(commands) {} 
    set ${this}(orient) horizontal 
    set ${this}(length) 400 
    
    # Private fields
    set ${this}(sliders) 0
    set ${this}(widget) $widget
    set ${this}(rect_width) 5
    set ${this}(rect_height) 10

    for { set i 0 } { $i < [llength $args] } { incr i } {
	set param [lindex $args $i]
	for { set j 0 } { $j < [llength $multiscaleParams] } { incr j } {
	    if { $param == "-[lindex [lindex $multiscaleParams $j] 0]" } {
		set field [lindex [lindex $multiscaleParams $j] 0]
		set lTypes [lindex [lindex $multiscaleParams $j] 1]
		for { set k 0 } { $k < [llength $lTypes] } { incr k } {
		    incr i
		    if { $i >= [llength $args] } {
			puts "multiscale: -$param missing argument $k of type [lindex $lTypes $k]"
			return 1
		    }
		    set ${this}($field) [lindex $args $i]
		}
		break
	    }
	}
	if { $j >= [llength $multiscaleParams] } {
	    puts "multiscale: bad argument -$param"
	    return 1
	}
    }

    ##################################################
    # Check the parameter values
    
    # sliders
    set ${this}(sliders) [llength [subst $${this}(variables)]]

    # variables
    if { [llength [subst $${this}(variables)]] < 1 } {
	puts "multiscale: too few variables: [llength [subst $${this}(variables)]] < 1"
	return 1
    }

    foreach var [subst $${this}(variables)] {
	regsub {\([_a-zA-Z0-9]*\)} $var "" var
	global $var
    }

    # commands
    set i 0
    foreach com [subst $${this}(commands)] {
	if { $i >= [subst $${this}(sliders)] } {
	    puts "multiscale: too many commands:  sliders = [subst $${this}(sliders)]"
	    return 1
	}
    }

    ##################################################
    # Create widget stuff
    
    set width 600
    set height 10

    if { [subst $${this}(length)] != {} && 0 < [subst $${this}(length)] } {
	set width [subst $${this}(length)]
    }

    # Build from 0 to 100, then scale and move it later.
    # Distance between left edges of boxes is 100.
    # The box is 10 wide, therefore the slider is 110 long.
    # The left box sticks up, and the right one hangs down.
    canvas $widget
    $widget create rect 0 0 [expr $width + [subst $${this}(rect_width)]] $height -fill grey -tag rail

    for { set i 0 } { $i < [subst $${this}(sliders)] } { incr i 1 } {
	set tag slider_$i
	set var [lindex [subst $${this}(variables)] $i]

	trace variable $var w "multiscaleVariableChanged $this $i"
	$widget create rect 0 0 [subst $${this}(rect_width)] [subst $${this}(rect_height)] -fill black -tags [list $tag ${tag}_rect]

	$widget create text 5 12 -anchor n -text [TruncToResolution [subst $$var] [subst $${this}(resolution)]] -fill black -tags [list $tag ${tag}_text]
    }

    # Move everything into view
    $widget move all $height $height
    for {set i 0} {$i < [subst $${this}(sliders)]} {incr i 1} {
	set tag slider_$i
	set var [lindex [subst $${this}(variables)] $i]

	$widget move $tag [expr [multiscaleValueToPixel $this [subst $$var]] - [lindex [$widget coords rail] 0]] 0
    }

    # Make the canvas fit comfortably around the image
    set bbox [$widget bbox all]
    set width [expr [lindex $bbox 2] + $height / 2]
    set height [expr [lindex $bbox 3] + $height / 2]
    $widget config -height $height -width $width

    # Bind drag actions 
    for {set i 0} {$i < [subst $${this}(sliders)]} {incr i 1} {
	$widget bind slider_$i <Button-1> [list multiscaleClick $this $i %x]
	$widget bind slider_$i <B1-Motion> [list multiscaleDrag $this $i %x]
    }

#    pack $widget

    return 0
}

proc multiscaleDisplay { name } {

    global multiscaleParams
    global multiscalePrivates

    # Create the associative array

    set this ""
    regsub -all "\\." multiscaleobject$name "_" this
    
    global $this

    puts "multiscale $name"
    puts "  public:"
    for { set i 0 } { $i < [llength $multiscaleParams] } { incr i } {
	puts "    -[lindex [lindex $multiscaleParams $i] 0] [subst $${this}([lindex [lindex $multiscaleParams $i] 0])]"
    }
    puts "  private:"
    for { set i 0 } { $i < [llength $multiscalePrivates] } { incr i } {
	puts "    -[lindex [lindex $multiscalePrivates $i] 0] [subst $${this}([lindex [lindex $multiscalePrivates $i] 0])]"
    }

}

######################################################################
# Test Code

# proc Display { var value } {
#     puts "Display: setting $var = $value"
# }

# set sliders 4
# set vars {}
# set commands {}
# for { set i 0 } { $i < $sliders } { incr i } {
#     set v$i [expr $i * 1.0 / ($sliders - 1)]
#     lappend vars v$i
#     lappend commands [list Display v$i]
# }

# puts "multiscale .ms -variables $vars -commands $commands"
# multiscale .ms -variables $vars -commands $commands -length 400 -spacing 0.1





